__author__ = "Syun'ichi Shiraiwa"
__copyright__ = "Copyright, S. Shiraiwa, PiScope Project"
__credits__ = ["Syun'ichi Shiraiwa"]
__version__ = "0.5"
__maintainer__ = "S. Shiraiwa"
__email__ = "shiraiwa@psfc.mit.edu"
__status__ = "beta"

from ifigure.mto.fig_obj import FigObj
from ifigure.mto.fig_page import FigPage
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.utils.cbook as cbook
import ifigure.utils.geom as geom
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
import logging
import matplotlib
from matplotlib.patches import Rectangle
import matplotlib.path as mpath

from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod

from ifigure.mto.generic_points import GenericPoint, GenericPointsHolder
from ifigure.mto.figobj_gpholder import FigObjGPHolder


class FigBox(FigObjGPHolder):
    def __init__(self, xy=[0, 0], width=0.1, height=0.1,
                 isotropic=False,
                 trans1=['figure']*2,
                 figaxes1=None,  **kywds):
        '''
        Curve object:
           figaxe1, figaxes2: None or full_path string
        '''

        self._objs = []  # for debug....
        if "draggable" in kywds:
            self.setvar("draggable", kywds["draggable"])
            del kywds["draggable"]
        else:
            self.setvar("draggable", True)
        args = ()

        GenericPointsHolder.__init__(self, num=2)
        super(FigBox, self).__init__(*args, **kywds)

        self.setvar("trans1", trans1)

        if figaxes1 is not None:
            self.setvar("figaxes1", figaxes1.get_full_path())
        else:
            self.setvar("figaxes1", None)
        self.setvar("width", width)
        self.setvar("height", height)
        self.setvar("xy", xy)
        self.setp('use_var', True)
        self.setp('isotropic', isotropic)

        self._cb_added = False
        self._drag_mode = 0
        self._drag_artist = None
        self._hit_seg_i = -1

    @classmethod
    def isFigBox(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return False

    @classmethod
    def get_namebase(self):
        return 'box'

    @classmethod
    def property_in_file(self):
        return ["facecolor", "edgecolor", "zorder", "alpha", "linewidth",
                "linestyle", "fill"]

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return super(FigBox, self).attr_in_file()

    @classmethod
    def property_in_palette(self):
        return ["path", "patch"], [["edgecolor", "linewidth",
                                    "plinestyle", "alpha", "box_type"],
                                   ["fill", "facecolor"]]

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'box.png')
        return [idx1]

    def get_containter(self):
        return self.get_figpage()._artists[0]

#    def set_parent(self, parent):
#        FigObj.set_parent(self, parent)
#        GenericPointsHolder.set_gp_figpage(self)

    def isDraggable(self):
        return self._var["draggable"]

    def generate_artist(self):

        if self.isempty() is False:
            return
        if self.getp('use_var'):
            trans1 = self.getvar("trans1")
            figaxes1 = self.getvar("figaxes1")
            if figaxes1 is not None:
                figaxes1 = self.find_by_full_path(str(figaxes1))

            xy = self.getvar("xy")
            w = self.getvar("width")
            h = self.getvar("height")

            self.set_gp_point(0, xy[0], xy[1], trans=trans1)
            self.set_gp_point(1, xy[0]+w, xy[1]+h, trans=trans1)
            self.set_gp_figpage()
            self.set_gp_figaxes(0, figaxes1)
            self.set_gp_figaxes(1, figaxes1)
            self.setp('use_var', False)

#        if self.hasp("loaded_property"):
#            lp = self.getp("loaded_property")
#            self.set_gp_point(0, 0, 0, trans = 'figure')
#            self.set_gp_point(1, 0, 0, trans = 'figure')
#            self.set_gp_figpage()
#            self.get_gp(0).dict_to_gp(lp[-1][0], self)
#            self.get_gp(1).dict_to_gp(lp[-1][1], self)

        a = self.make_newartist()
        if self.hasp("loaded_property"):
            self.set_artist_property(a, lp[0])
            self.delp("loaded_property")
        self.add_artists(a)
        if not self._cb_added:
            fig_page = self.get_figpage()
            fig_page.add_resize_cb(self)
            self._cb_added = True

    def onResize(self, evt):
        self.refresh_artist()

    def do_update_artist(self):
        self.refresh_artist()

    def make_newartist(self):
        self.check_loaded_gp_data()
        x1, y1 = self.get_gp(0).get_device_point()
        x2, y2 = self.get_gp(1).get_device_point()

        xy = (min([x1, x2]), min([y1, y2]))
        w = abs(x1-x2)
        h = abs(y1-y2)
        if self.getp('isotropic'):
            if w < h:
                h = w
            else:
                w = h

        a = Rectangle(xy, w, h, facecolor='none', fill=False,
                      edgecolor='black', alpha=1)

        lp = self.getp("loaded_property")
        if lp is not None:
            self.set_artist_property(a, lp[0])
            self.delp("loaded_property")
        a.figobj = self
        a.figobj_hl = []

        a.set_zorder(self.getp('zorder'))
        a._is_frameart = self.getp('frameart')
        return a

    def add_artists(self, a):
        figure = self.get_containter()
        self._artists = [a]
        figure.patches.append(a)
        a.set_figure(figure)

    def refresh_artist(self):
        if len(self._artists) != 1:
            return
        a1 = self._artists[0]
        z = a1.zorder
        hl = len(self._artists[0].figobj_hl) != 0
        self.del_artist(delall=True)
        a = self.make_newartist()
        a.set_zorder(z)
        self.add_artists(a)
        self.highlight_artist(hl, [a])
        a2 = self._artists[0]
        if a1 != a2:
            ifigure.events.SendReplaceEvent(self, a1=a1, a2=a2)

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = artist

        # save_data2->load_data2 will set "loaded_property"

        self.load_data2(self.save_data2({}))

#        for a in self._artists:
#           val.append(self.get_artist_property(a))
#        if len(val) != 0:
#           self.setp("loaded_property", val)

        if len(artistlist) != 0:
            container = self.get_figpage()._artists[0]
            self.highlight_artist(False, artistlist)
            for a in artistlist:
                #            a is axes in this case
                #            a.remove()
                container.patches.remove(a)

        super(FigBox, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist
        figure = self.get_figpage()._artists[0]
        if val == True:
            for a in alist:
                box = a.get_window_extent().get_points()
                x = [box[0][0], box[0][0], box[1][0], box[1][0], box[0][0]]
                y = [box[0][1], box[1][1], box[1][1], box[0][1], box[0][1]]
                hl = matplotlib.lines.Line2D(x, y, marker='s',
                                             color='k', linestyle='None',
                                             markerfacecolor='none',
                                             markeredgewidth=0.5,
                                             figure=figure)
                figure.lines.extend([hl])
                a.figobj_hl.append(hl)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    figure.lines.remove(hl)
                    a.figobj_hl = []

    def get_artist_extent(self, a):
        box = a.get_window_extent().get_points()
        return [box[0][0], box[1][0], box[0][1], box[1][1]]

    def picker_a0(self, a, evt):

        self._picker_a_mode = 0
        hit, extra, type, loc = super(FigBox, self).picker_a0(a, evt)
        return hit, extra, type, loc

    def drag_a(self, a, evt, shift=None, scale=None):
        redraw, scale = super(FigBox, self).drag_a(
            a, evt, shift=shift, scale=scale)
        return redraw, scale

    def dragdone_a(self, a, evt, shift=None, scale=None):
        shift = evt.guiEvent_memory.ShiftDown()
        redraw, scale0 = super(FigBox, self).dragdone_a(a, evt,
                                                        shift=shift, scale=scale)

        window = evt.guiEvent_memory.GetEventObject().GetTopLevelParent()
        hist = GlobalHistory().get_history(window)

        h = []
        h = h + self.scale_artist(scale0, action=True)

        hist.start_record()
        for item in h:
            hist.add_history(item)
        hist.stop_record()
        return 0, scale0

    def scale_artist(self, scale, action=True):
        st_extent = self.get_artist_extent(self._artists[0])
        rec = geom.scale_rect(st_extent,
                              scale)
        gp = self.get_gp(0)
        x1, y1 = gp.get_device_point()
        gp = self.get_gp(1)
        x2, y2 = gp.get_device_point()

        if x1 > x2 and y1 > y2:
            i = (1, 3, 0, 2)
        elif x1 <= x2 and y1 > y2:
            i = (0, 3, 1, 2)
        elif x1 > x2 and y1 <= y2:
            i = (1, 2, 0, 3)
        elif x1 <= x2 and y1 <= y2:
            i = (0, 2, 1, 3)
        dx1 = rec[i[0]] - st_extent[i[0]]
        dy1 = rec[i[1]] - st_extent[i[1]]
        dx2 = rec[i[2]] - st_extent[i[2]]
        dy2 = rec[i[3]] - st_extent[i[3]]

        if action:
            h = []
            a1 = self.move_gp_points(0, dx1, dy1, action=True)
            a2 = self.move_gp_points(1, dx2, dy2, action=True)
            h.append(a1)
            h.append(a2)
            return h
        else:
            self.move_gp_points(0, dx1, dy1, action=False)
            self.move_gp_points(1, dx2, dy2, action=False)

    def canvas_menu(self):
        return self.gp_holder_canvas_menu()

    def get_boxtype(self, a):
        if self.getp('isotropic'):
            return 'square'
        else:
            return 'rectangle'

    def set_boxtype(self, value, a):
        if value == 'square':
            self.setp('isotropic', True)
        else:
            self.setp('isotropic', False)
        self.refresh_artist()



from mpl_toolkits.mplot3d.axis3d import XAxis as _XAxis
from mpl_toolkits.mplot3d.axis3d import YAxis as _YAxis
from mpl_toolkits.mplot3d.axis3d import ZAxis as _ZAxis


class XAxis(_XAxis):
    def set_ticks_position(self, position):
        pass
class YAxis(_YAxis):
    def set_ticks_position(self, position):
        pass
class ZAxis(_ZAxis):
    def set_ticks_position(self, position):
        pass



import sys
import socket
import shlex
import threading
import os
import binascii
import subprocess
import time
import tempfile
from six.moves import socketserver
import traceback
import ifigure.utils.pickle_wrapper as pickle
from ifigure.utils.pickled_pipe import PickledPipe

from ifigure.mdsplus.proxy_jobrunner import JobRunner

if __name__ == '__main__':
    host = sys.argv[1]
    port = sys.argv[2]
    ch = PickledPipe(sys.stdin, sys.stdout)
    runner = JobRunner(host, port)
    while True:
        time.sleep(0.01)
        try:
            jobset = ch.recv(nowait=False)
        except EOFError:
            r = {'error message': ['message was received']}
            r['worker output'] = ''
            ch.send(r)

        if jobset != -1:
            tmp = tempfile.TemporaryFile()
            sys.stdout = tmp
            try:
                r = runner.run(jobset)
            except:
                r = {'error message': [
                    'runner.run failed', traceback.format_exc()]}
            finally:
                tmp.flush()
                tmp.seek(0)
                r['worker output'] = tmp.read()
                tmp.close()
                jobs, names = jobset
                def_names = ['message',
                             'pickle error',
                             'network message',
                             'worker output',
                             'error message',
                             'x', 'y', 'z',
                             'xerr', 'yerr', 'zerr']
                for n in names:
                    if not n in def_names:
                        def_names.append(n)
                r2 = {}
                for n in def_names:
                    if n in r:
                        r2[n] = r[n]
                ch.send(r2)
        else:
            runner.terminate()
            ch.send('ok')
            break



# Very minimalist password entry dialogue
import wx
from ifigure.utils.wx3to4 import FlexGridSizer


class UsernamePasswordDialog(wx.Dialog):
    def __init__(self, parent, id=-1, title="Enter password", label="Enter Password;"):
        wx.Dialog.__init__(self, parent, id, title,
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.label = wx.StaticText(self, label=label)

        sizer = FlexGridSizer(2, 2)
        sizer.AddGrowableCol(1, 1)
        st1 = wx.StaticText(self, label="Name")
        self.field = wx.TextCtrl(self, value="", size=(-1, -1))
        st2 = wx.StaticText(self, label="Password")
        self.field2 = wx.TextCtrl(
            self, value="", size=(-1, -1), style=wx.TE_PASSWORD)
        sizer.Add(st1, 0, wx.ALL, 2)
        sizer.Add(self.field, 1, wx.ALL | wx.EXPAND, 2)
        sizer.Add(st2, 0, wx.ALL, 2)
        sizer.Add(self.field2, 1, wx.ALL | wx.EXPAND, 2)

        self.okbutton = wx.Button(self, label="OK", id=wx.ID_OK)
        self.cancelbutton = wx.Button(self, label="Cancel", id=wx.ID_CANCEL)
        self.mainSizer.Add(self.label, 0, wx.ALL, 8)
        self.mainSizer.Add(sizer, 1, wx.ALL | wx.EXPAND, 2)
        self.buttonSizer.Add(self.okbutton, 0, wx.ALL, 8)
        self.buttonSizer.Add(self.cancelbutton, 0, wx.ALL, 8)
        self.mainSizer.Add(self.buttonSizer, 0, wx.ALL, 0)
        self.Bind(wx.EVT_BUTTON, self.onOK, id=wx.ID_OK)
        self.Bind(wx.EVT_BUTTON, self.onCancel, id=wx.ID_CANCEL)
        #self.Bind(wx.EVT_TEXT_ENTER, self.onOK)
        self.SetSizer(self.mainSizer)
        self.Fit()
        self.Centre()
        self.result = None

#    def onClose(self):

    def onOK(self, event):
        self.result = [self.field.GetValue(), self.field2.GetValue()]
        if self.IsModal():
            self.EndModal(wx.ID_OK)
        else:
            self.Destroy()

    def onCancel(self, event):
        if self.IsModal():
            self.EndModal(wx.ID_CANCEL)
        else:
            self.Destroy()


class PasswordDialog(wx.Dialog):
    def __init__(self, parent, id=-1, title="Enter password", label="Enter Password;"):
        wx.Dialog.__init__(self, parent, id, title)
        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.label = wx.StaticText(self, label=label)
        self.field = wx.TextCtrl(self, value="", size=(
            300, -1), style=wx.TE_PASSWORD | wx.TE_PROCESS_ENTER)
        self.okbutton = wx.Button(self, label="OK", id=wx.ID_OK)
        self.cancelbutton = wx.Button(self, label="Cancel", id=wx.ID_CANCEL)
        self.mainSizer.Add(self.label, 0, wx.ALL, 8)
        self.mainSizer.Add(self.field, 0, wx.ALL, 8)
        self.buttonSizer.Add(self.okbutton, 0, wx.ALL, 8)
        self.buttonSizer.Add(self.cancelbutton, 0, wx.ALL, 8)
        self.mainSizer.Add(self.buttonSizer, 0, wx.ALL, 0)
        self.Bind(wx.EVT_BUTTON, self.onOK, id=wx.ID_OK)
        self.Bind(wx.EVT_BUTTON, self.onCancel, id=wx.ID_CANCEL)
        self.Bind(wx.EVT_TEXT_ENTER, self.onOK)
        self.SetSizer(self.mainSizer)
        self.Fit()
        self.result = None

    def onOK(self, event):
        self.result = self.field.GetValue()
        if self.IsModal():
            self.EndModal(wx.ID_OK)
        else:
            self.Destroy()

    def onCancel(self, event):
        if self.IsModal():
            self.EndModal(wx.ID_CANCEL)
        else:
            self.Destroy()



from __future__ import print_function
import numpy as np
import scipy.io as sio
import wx
import os
import traceback
import shutil
import ifigure
from ifigure.mto.py_contents import Matfile, MatData
######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.
#      Strong recommendation : make module "independent".
#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  module loading
#      order-dependency problem.
#
#   name of module/class
module_name = 'matlab_matfile'
class_name = 'matlab_matfile'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
menu = [("Import...",   "onLoadFile", True),
        ("Export...",   "onExportFile", True),
        ("Update File",   "onUpdateFile", True),
        ("Update Tree",   "onUpdateTree", True)]
method = ['onLoadFile', 'onExportFile',
          'onUpdateFile', 'onUpdateTree',
          'init', 'init_after_load']

icon = 'data.png'


modename = 'mat_pathmode'
pathname = 'mat_path'
extname = 'mat_ext'
wildcard = 'mat(*.mat)|*.mat|Any|*'
######## Do not chage this field ####################
can_have_child = False
has_private_owndir = False
nosave_var0 = True
######################################################


def init(self, *args, **kargs):
    #   a function called when py_module is initialized
    obj = self.td
    obj.mk_owndir()
    nm = Matfile()
    nm['data'] = MatData()
    obj.setvar0(nm)
    if 'src' not in kargs:
        self.onLoadFile(None)
    # self.onLoadFile()


def load_matfile(obj):
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)
    if file == '':
        return
    # clean first
    for name, child in obj.get_children():
        child.destroy()

    nm0 = sio.loadmat(file)
    nm = Matfile()
    nm['data'] = MatData()
    obj.setvar0(nm)
    mtime = os.path.getmtime(file)
    obj.setvar('mat_mtime', mtime)

    if nm0 is None:
        return
    for key in nm0:
        nm['data'][key] = nm0[key]


def export_matfile(obj, file):
    nm = obj.getvar0()
    nm = sio.savemat(file, nm['data'])


def onUpdateFile(self, e):
    obj = self.td
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)
    file_check = os.path.join(obj.owndir(), obj.name)+'.mat'

    if file == '':
        file = file_check
        try:
            export_matfile(obj, file)
            obj.set_path_pathmode(file, modename, pathname, extname)
        except:
            print('file save failed')
            traceback.print_exc()
    else:
        if file_check != file:
            shutil.move(file, file_check)
            obj.set_path_pathmode(file_check, modename, pathname, extname)
            file = file_check
        try:
            export_matfile(obj, file)
        except:
            print('file save failed')
            traceback.print_exc()


def onUpdateTree(self, e):
    obj = self.td
    load_matfile(obj)


def onLoadFile(self, e=None, file='', checklist=None):
    from ifigure.utils.addon_utils import onLoadFile
    if file != '':
        self.td.set_path_pathmode(file, modename, pathname, extname,
                                  checklist=checklist)
        ret = True
    else:
        ret = onLoadFile(self.td, message="Select .mat File",
                         modename=modename,
                         pathname=pathname,
                         extname=extname,
                         wildcard=wildcard,
                         ask_org_copy=True)
    if ret:
        load_matfile(self.td)
        ifigure.events.SendChangedEvent(self.td)


def onExportFile(self, e):
    from ifigure.utils.addon_utils import onWriteFile
    app = wx.GetApp().TopWindow
    obj = self.td
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)

    open_dlg = wx.FileDialog(app,
                             message='Save .mat file',
                             defaultDir=os.path.dirname(file),
                             defaultFile='untitled.mat',
                             style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT,
                             wildcard=wildcard)

    if open_dlg.ShowModal() != wx.ID_OK:
        open_dlg.Destroy()
        return
    else:
        filename = str(open_dlg.GetPath())
        open_dlg.Destroy()
    export_matfile(obj, filename)


def init_after_load(self, olist, nlist):
    obj = self.td
    load_matfile(obj)



from ifigure.mto.triangle_plots import TrianglePlots
from matplotlib.collections import LineCollection
from matplotlib.cm import ScalarMappable

from matplotlib.artist import allow_rasterization
from ifigure.mto.fig_obj import FigObj
from ifigure.mto.axis_user import XUser, YUser, ZUser, CUser
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
import logging
from ifigure.utils.cbook import ProcessKeywords, LoadImageFile, isdynamic, isiterable
from ifigure.utils.triangulation_wrapper import tri_args
from ifigure.matplotlib_mod.triplot_mod import triplot
from ifigure.utils.args_parser import ArgsParser
from matplotlib.path import Path
from matplotlib.transforms import Bbox, TransformedPath
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigPlot')


class FigTriplot(FigObj, XUser, YUser, TrianglePlots):
    '''
    triplot(x, y)
    triplot(tri, x, y)
    '''
    def __new__(cls, *args, **kywds):
        if len(args) == 3:
            tri = args[0]
            args = args[1:]
        else:
            tri = None

        def set_hidden_vars(obj):
            obj._other_artists = []
            obj._objs = []  # for debug....
            obj._data_extent = None
            return obj
        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj
        p = ArgsParser()
        p.add_opt('x', None, ['numbers|nonstr', 'dynamic'])
        p.add_var('y', ['numbers|nonstr', 'dynamic'])
        p.add_key('color', 'b')
        p.add_key('linestyle', '-')
        p.add_key('linewidth', 1.0)
        p.set_ndconvert("x", "y")
        p.set_squeeze_minimum_1D("x", "y")
        p.add_key('mask', None)

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)
        if (v["x"] is None and not isdynamic(v["y"])):
            v["x"] = np.arange(v["y"].shape[-1])
        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        obj.setvar("tri", tri)

        return obj

    def __init__(self, *args,  **kywds):
        if len(args) == 3:
            tri = args[0]
            args = args[1:]
        else:
            tri = None
        self._tri = tri
        XUser.__init__(self)
        YUser.__init__(self)
        TrianglePlots.__init__(self)
        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigTriplot, self).__init__(*args, **kywds)

    @classmethod
    def isFigTriplot(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return True

    @classmethod
    def get_namebase(self):
        return 'triplot'

    def property_in_file(self):
        return (["color", "linestyle",
                 "linewidth", "marker", "markersize",
                 "markeredgecolor", "markerfacecolor",
                 "markeredgewidth", "alpha"] +
                super(FigTriplot, self).property_in_file())

    @classmethod
    def property_in_palette(self):
        return (["line", "marker"],
                [["color_2", "plinestyle_2", "linewidth_2", "alpha_2", ],
                 ["markerfacecolor", "markeredgecolor", "marker",
                  "markersize", "markeredgewidth"], ])

    @classmethod
    def attr_in_file(self):
        return (['mask', 'color', 'linestyle', 'linewidth'] +
                super(FigTriplot, self).attr_in_file())

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'plot1.png')
        return [idx1]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)

        super(FigTriplot, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)

    def args2var(self):
        names0 = self.attr_in_file()
        names = ["x", "y"]
        use_np = [True]*2 + [False]
        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)
        x = values[0]
        y = values[1]
        if x is None:
            x = np.arange((y.shape)[-1])
            self.setp("x", x)
        if x is None:
            return False
        if y is None:
            return False
        if (x.ndim == 1 and y.ndim == 1 and
                x.size != y.size):
            self.setp("x", None)
            self.setp("y", None)
            return False
        #self.setp('x', self.getp('x').flatten())
        #self.setp('y', self.getp('y').flatten())

        return True

    def generate_artist(self):
        container = self.get_container()
        if self.isempty() is False:
            return

        x, y = self._eval_xy()  # this handles "use_var"

        lp = self.getp("loaded_property")

        if True:
            x, y = self.getp(("x", "y"))
            if y is None:
                return
            if x is None:
                return

            if (x is not None and
                    y is not None):
                self._data_extent = [np.nanmin(x), np.nanmax(x),
                                     np.nanmin(y), np.nanmax(y)]

                # if len(y.shape) == 1:
                if True:
                    kywds = self._var["kywds"]
                    args, self._tri = tri_args(x, y, self._tri)
                    kywds['mask'] = self.getp('mask')
                    kywds['linestyle'] = self.getp('linestyle')
                    kywds['linewidth'] = self.getp('linewidth')
                    kywds['color'] = self.getp('color')
                    a = triplot(container, *args, **kywds)
                    self.set_artist(a[0])
                    self._other_artists = a[1:]

        if lp is not None:
            for i in range(0, len(lp)):
                self.set_artist_property(self._artists[i], lp[i])
            self.delp("loaded_property")
        self.set_rasterized()

    def refresh_artist_data(self):
        if self.isempty() and not self._suppress:
            self.generate_artist()
            return
        self._data_extent = None
        x, y = self._eval_xy()  # this handles "use_var"
        self._artists[0].set_xdata(x)
        self._artists[0].set_ydata(y)

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = artist

        self.store_loaded_property()
        if len(artistlist) is len(self._artists):
            for name, child in self.get_children():
                child.del_artist(delall=True)
        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            for a in artistlist:
                #             a.set_picker(None)
                try:
                    if a in container.collections:
                        a.remove()
                    elif a in container.lines:
                        a.remove()
                except:
                    logging.exception("fig_plot: highlight_aritst() failed")
            for a in self._other_artists:
                a.remove()

        super(FigTriplot, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist

        container = self.get_container()
        if val == True:
            for a in alist:
                x = a.get_xdata()
                y = a.get_ydata()
                p = 300
                if x.shape[0] > p:
                    idx = np.floor(
                        np.arange(p)*(x.shape[0]/float(p))).astype(int)
                else:
                    idx = np.arange(x.shape[0])
                hl = container.plot(x[idx].copy(), y[idx].copy(), marker='s',
                                    color='k', linestyle='None',
                                    markerfacecolor='None',
                                    markeredgewidth=0.5,
                                    scalex=False, scaley=False)

                for item in hl:
                    a.figobj_hl.append(item)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    try:
                        hl.remove()
                    except:
                        logging.exception(
                            "fig_triplot: highlight_aritst() failed")
                a.figobj_hl = []
#
#   def hit_test
#

    def picker_a(self, artist, evt):
        axes = artist.axes
        if axes is None:
            return False, {}

        hit, extra = artist.contains(evt)

        if hit:
            return True, {"child_artist": artist}
        else:
            return False, {}

    def picker_a0(self, artist, evt):
        hit, extra = self.picker_a(artist, evt)
        if hit:
            return hit, extra, 'area', 3
        else:
            return False, {}, None, 0
#
#  Popup in Canvas
#

    def onExport(self, event):
        from matplotlib.artist import getp

        shell = self.get_root_parent().app.shell
        canvas = event.GetEventObject()
        sel = [a() for a in canvas.selection]
        for a in self._artists:
            if a in sel:
                fig_val = {"xdata": getp(a, "xdata"),
                           "ydata": getp(a, "ydata")}
                text = '#Exporting data as fig_val[\'xdata\'], fig_val[\'ydata\']\"'
                self._export_shell(fig_val, 'fig_val', text)
                break
#
#   setter/getter
#

    def set_color(self, value, a):
        self.setp('color', value)
        self._setter_others(value, 'color')

    def get_color(self, a):
        return self.getp('color')

    def set_linewidth(self, value, a):
        self._setter_others(value, 'linewidth')

    def get_linewidth(self, a):
        self._getter_others('linewidth')

    def set_linestyle(self, value, a):
        self._setter_others(value, 'linestyle')

    def get_linestyle(self, a):
        self._getter_others('linestyle')

    def set_alpha(self, value, a):
        self._setter_others(value, 'alpha')
        for a in self._artists:
            a.set_alpha(value)

    def get_alpha(self, a):
        return self._getter_others('alpha')

    def _setter_others(self, value, prop):
        for a in self._other_artists:
            m = getattr(a, 'set_'+prop)
            m(value)

    def _getter_others(self, prop):
        a = self._other_artists[0]
        m = getattr(a, 'get_'+prop)
        return m()

#
#   data extent
#
    def get_data_extent(self):
        if self._data_extent is not None:
            return self._data_extent
        x, y = self._eval_xy()  # this handles "use_var"
        if self.isempty():
            if x is None:
                self._data_extent = [0, len(y), np.nanmin(y), np.nanmax(y)]
            else:
                self._data_extent = [np.nanmin(x), np.nanmax(
                    x), np.nanmin(y), np.nanmax(y)]
        else:
            xr = (np.inf, -np.inf)
            yr = (np.inf, -np.inf)
            for a in self._artists:
                x = a.get_xdata()
                y = a.get_ydata()
                xr = (min((xr[0], np.nanmin(x))), max((xr[1], np.nanmax(x))))
                yr = (min((yr[0], np.nanmin(y))), max((yr[1], np.nanmax(y))))
            self._data_extent = [xr[0], xr[1], yr[0], yr[1]]
        return self._data_extent

#
#   range
#
    def get_xrange(self, xrange=[None, None], scale='linear'):
        x, y = self._eval_xy()
        if x is None:
            return
        if scale == 'log':
            x = mask_negative(x)
        return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        def calc_yrange(x, y, xrange, yrange, scale):
            # x.ndim = 1, y.ndim = 1 in this func
            idx = np.where((x >= xrange[0]) & (x <= xrange[1]))
            check = False
            for ar in idx:
                if len(ar) == 0:
                    check = True
            if check:
                return yrange

            yy = y[idx]
            if scale == 'log':
                yy = mask_negative(yy)
            y0n = np.nanmin(yy)
            y1n = np.nanmax(yy)
            yrange = self._update_range(yrange, (y0n, y1n))
            return yrange

        de = self.get_data_extent()
        y0 = yrange[0]
        y1 = yrange[1]

        if (xrange[0] is not None and
                xrange[1] is not None):
            x, y = self._eval_xy()

            if x.ndim > 2 or y.ndim > 2:
                return yrange
            if x.ndim == 1:
                xl = 1
            else:
                xl = x.shape[0]
            if y.ndim == 1:
                yl = 1
            else:
                yl = y.shape[0]
            for k in range(xl):
                if x.ndim == 1:
                    xt = x
                else:
                    xt = x[k]
                for m in range(yl):
                    if y.ndim == 1:
                        yt = y
                    else:
                        yt = y[m]
                    yrange = calc_yrange(xt, yt, xrange, yrange, scale)

        if (not (xrange[0] is not None and
                 xrange[1] is not None) or
                (yrange[0] is None) or (yrange[1] is None)):
            if de is not None:
                yrange = self._update_range(yrange, de[2:])
        return yrange

    def _saveload_names(self):
        return ['x', 'y']

    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        names = self._saveload_names()
        var = {name: check(self, name) for name in names}
        for name in names:
            if not var[name]:
                if self._save_mode == 0 or name == 's':
                    var[name+'data'] = self.getp(name)
                else:
                    var[name+'data'] = np.array([0, 0])
        data['FigTriplot'] = (1, var)
        data = super(FigTriplot, self).save_data2(data)

        return data

    def prepare_compact_savemode(self):
        var_bk = self._var.copy()
        self._var['x'] = np.array([0, 0])
        self._var['y'] = np.array([0, 0])
        return var_bk

    def load_data2(self, data):
        d = data['FigTriplot']
        super(FigTriplot, self).load_data2(data)
        var = d[1]
        names = self._saveload_names()
        for name in names:
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def _eval_xy(self):
        names = ("x", "y")
        if self._tri is None:
            self._tri = self.getvar('tri')
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return [None]*len(names)
        return self.getp(names)



from __future__ import print_function
import multiprocessing as mp
import threading
from six.moves import queue as Queue
import traceback
import time
from weakref import WeakKeyDictionary
import __future__

try:
    use_event_listener = True
    import MDSplus
except ImportError:
    use_event_listener = False
use_event_listener = True


class EventListenerProc(mp.Process):
    class EventThread(MDSplus.Event):
        def __init__(self, queue, event_name,  *args, **kargs):
            self.queue = queue
            self.event_name = event_name
            Event.__init__(self, event_name, *args, **kargs)

        def run(self):

            self.queue.put(('send', self.event_name))

    def __init__(self, task_queue, result_queue, *args, **kargs):
        #   def __init__(self,  *args, **kargs):
        super(EventListenerProc, self).__init__(*args, **kargs)
        self.task_queue = task_queue
        #self.sub_queue = Queue.Queue()
        self.result_queue = result_queue
        self.event_threads = []
        self.event_list = []

    def run(self, *args, **kargs):
        while True:
            #            while True:
            try:
                jobcode, jobparam = self.task_queue.get(True)  # True, 0.1)

            except:
                continue
                pass

            # jobset
            ###   (jobcode, jobparam)
            # jobcode : 'wait' : launch thread if it is not waiting
            # jobcode : 'exit' : exit loop
            # jobcode : 'send' :  when event is recieved
            if jobcode == 'wait':
                # jobparam = event name
                if not jobparam in self.event_list:
                    o = EventThread(self.task_queue, jobparam)
                    self.event_threads.append(o)
                    self.event_list.append(jobparam)
            elif jobcode == 'send':
                # jobparam = event thread
                self.result_queue.put(jobparam)
            elif jobcode == 'exit':
                for o in self.event_threads:
                    try:
                        o.cancel()
                    except:
                        traceback.print_exc()
                        print('MDSplus event cancel failed')
                        pass
                break
            elif jobcode == 'disconnect':
                if jobparam in self.event_list:
                    o = EventThread(self.task_queue, jobparam)
                    for t in self.event_threads:
                        if t.event_name == jobparam:
                            t.cancel()
            else:
                pass
#            self.task_queue.task_done()



from __future__ import print_function
import hgapi
import os
import traceback


def pull(dir1, dir2, dry_run=False):
    '''
       pull from dir2 to dir1

       example) pull('/Users/shiraiwa/hg_root', 'ssh://cmodws60/hg_root', dry_run = False)
    '''
    for root, dirs, files in os.walk(dir1):
        if '.hg' in dirs:
            dirs.remove('.hg')
            repo = hgapi.Repo(root)
            url = dir2+root[len(dir1):]
            try:
                out = repo.hg_command('incoming', url)
                lines = out.split('\n')
                numc = len([1 for l in lines if l.startswith('changeset')])
                if numc > 0:
                    print(str(numc) + ' is coming from ' + url)

            except:
                if traceback.format_exc().find('no changes found') != -1:
                    print(url + ' is updated')
                    continue
                if traceback.format_exc().find('There is no Mercurial') != -1:
                    print('no repo found in '+url)
                    continue
                traceback.print_exc()
                continue
            if dry_run:
                continue
            try:
                repo.hg_pull(url)
            except:
                traceback.print_exc()


def push(dir1, dir2, dry_run=False):
    '''
       pull from dir2 to dir1

       example) pull('/Users/shiraiwa/hg_root', 'ssh://cmodws60/hg_root', dry_run = False)
    '''
    for root, dirs, files in os.walk(dir1):
        if '.hg' in dirs:
            dirs.remove('.hg')
            repo = hgapi.Repo(root)
            url = dir2+root[len(dir1):]
            try:
                out = repo.hg_command('outgoing', url)
                lines = out.split('\n')
                numc = len([1 for l in lines if l.startswith('changeset')])
                if numc > 0:
                    print(str(numc) + ' is coming from ' + url)

            except:
                if traceback.format_exc().find('no changes found') != -1:
                    print(url + ' is updated')
                    continue
                if traceback.format_exc().find('There is no Mercurial') != -1:
                    print('no repo found in '+url)
                    continue
                traceback.print_exc()
                continue
            if dry_run:
                continue
            try:
                repo.hg_push(url)
            except:
                traceback.print_exc()



#
#   fig_hist
#
#       object to support histgram
#
#       note that this object itself does not support
#       multiple dataset. Multiple dataset is handled
#       in book_viewer_interactive::hist
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************

from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.mto.axis_user import XUser, YUser, CUser, ZUser
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import sys
import weakref
import numpy as np
import ifigure.utils.cbook as cbook
import ifigure.widgets.canvas.custom_picker as cpicker
from ifigure.utils.cbook import ProcessKeywords
from matplotlib.cm import ScalarMappable
from matplotlib.patches import Rectangle
from matplotlib.transforms import Bbox
from ifigure.utils.args_parser import ArgsParser
from matplotlib.colors import Colormap
from matplotlib.patches import Rectangle
from ifigure.widgets.undo_redo_history import UndoRedoAddRemoveArtists, GlobalHistory

#
#  debug setting
#
import ifigure.utils.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('FigHist')

default_kargs = {}


class FigHist(FigObj, XUser, YUser, ZUser):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._tri = None
            for key in default_kargs:
                if not obj.hasp(key):
                    obj.setp(key, default_kargs[key])
                if not obj.hasvar(key):
                    obj.setvar(key, default_kargs[key])
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        p = ArgsParser()
        p.add_var('x', ['any'])
#        p.set_default_list(default_kargs)
        p.add_key('alpha', 1.0)
        p.add_key('color', 'b')

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)

        for name in ('x', 'alpha', 'color'):
            obj.setvar(name, v[name])

        obj.setvar("kywds", kywds)

        return obj

    def __init__(self, *args, **kywds):
        self._data_extent = None
        XUser.__init__(self)
        YUser.__init__(self)
        ZUser.__init__(self)

        self._pick_pos = None
        self._cb_added = False
        self._n = None
        self._bins = None
        self._hit_a = None
        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigHist, self).__init__(**kywds)

    @classmethod
    def isFigHist(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'hist'

    @classmethod
    def property_in_file(self):
        return (['color', 'alpha', 'linestyle', 'linewidth'] +
                super(FigHist, self).property_in_file())

    @classmethod
    def property_in_palette(self):
        return ["color_2", "linewidth_2", "plinestyle_2",  "alpha_2"]

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return ([] +
                super(FigHist, self).attr_in_file())

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'image.png')
        return [idx1]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)
        ZUser.unset_az(self)
        super(FigHist, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)
        ZUser.get_zaxisparam(self)

    '''    
    def _args2var(self):
        names0 = self.attr_in_file()
        names  = ["x","y", "z"] + names0
        use_np = [True]*3 + [False]*len(names0)
        values = self.put_args2var(names,
                                   use_np) 
        x, y, z, = values[:3]
        if x is None: return False
        if y is None: return False

        return True
    '''

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change
        container = self.get_container()
        if container is None:
            return
#           if self.get_figaxes().get_3d(): return
        if self.isempty() is False:
            return
        lp = self.getp("loaded_property")
        x = args = self.getvar('x')
        if lp is None or len(lp) == 0:

            if self.get_figaxes().get_3d():
                args = (x,)
            else:
                args = (x,)

            kywds = self._var["kywds"].copy()
            kywds['alpha'] = self.getvar('alpha')
            kywds['color'] = self.getvar('color')
            n, bins, patches = container.hist(*args,
                                              **kywds)
            self.set_artist(patches)
        else:
            if self.get_figaxes().get_3d():
                args = (x,)
            else:
                args = (x,)
            kywds = lp[0]
            '''
               kywds['s'] = self.getvar('s')*self.getvar('sscale')               
               kywds['c'] = self.getvar('c')
               kywds['marker']  = self.getvar('marker')                              
               '''
            n, bins, patches = container.hist(*args,
                                              **kywds)
            self.set_artist(patches)
        self._n = n
        self._bins = bins
        self.delp("loaded_property")

    def call_adjust_range(self):
        parent = self.get_figaxes()
        if parent is not None:
            parent.adjust_axes_range()

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = self._artists

        self.store_loaded_property()
#        self.highlight_artist(False, artistlist)
        for a in artistlist:
            a.remove()

        super(FigHist, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist

        container = self.get_container()
        figure = self.get_figpage()._artists[0]

        if val == True:
            x = self.getvar('x')
            if self.get_figaxes().get_3d():
                args = (x,)
            else:
                args = (x, )

            x0, y0 = container.transAxes.transform((0, 0))
            x1, y1 = container.transAxes.transform((1, 1))
            bbox = Bbox([[x0, y0], [x1, y1]])

            for a in alist:
                hl = [Rectangle(p.get_xy(),
                                p.get_width(),
                                p.get_height(),
                                facecolor='black',
                                fill=True,
                                alpha=0.5,
                                figure=figure,
                                transform=container.transData)
                      for p in self._artists]
                a.figobj_hl = hl
                for aa in hl:
                    aa.set_clip_box(bbox)
                    aa.set_clip_on(True)
                    figure.patches.append(aa)
        else:
            for a in self._artists:
                for hl in a.figobj_hl:
                    if hl in figure.patches:
                        figure.patches.remove(hl)
                a.figobj_hl = []

    def picker_a(self, artist, evt):
        hit, extra = artist.contains(evt)
        if hit:
            self._hit_a = artist
            return True,  {'child_artist': artist}
        else:
            self._hit_a = None
            return False, {}

    def canvas_menu(self):
        if self._hit_a is None:
            return super(FigHist, self).tree_viewer_menu()
        menus = ([("Generate patch",   self.onGeneratePatch, None), ] +
                 super(FigHist, self).canvas_menu())
        return menus

    def onGeneratePatch(self, event):
        from ifigure.mto.fig_fill import FigFill
        if self._hit_a is not None:
            canvas = event.GetEventObject()
            x, y = self._hit_a.get_xy()
            w = self._hit_a.get_width()
            h = self._hit_a.get_height()

            x = [x, x+w, x+w, x, x]
            y = [y, y, y+h, y+h, y]
            obj = FigFill(x, y, color=self.getvar('color'))
            ax = self.get_figaxes()
            name = ax.get_next_name('HistBin')
            ax.add_child(name, obj)
            obj.realize()
            ax.set_bmp_update(False)
            canvas.draw()

            artists = [weakref.ref(obj._artists[0])]
            h = [UndoRedoAddRemoveArtists(artists=artists,
                                          mode=0)]
            window = canvas.GetTopLevelParent()
            GlobalHistory().get_history(window).make_entry(h, menu_name='add histbin')

    def set_linewidth(self, value, a=None):
        for a in self._artists:
            a.set_linewidth(value)

    def get_linewidth(self, a=None):
        v = self._artists[0].get_linewidth()
        return v

    def set_linestyle(self, value, a=None):
        for a in self._artists:
            a.set_linestyle(value)

    def get_linestyle(self, a):
        v = self._artists[0].get_linestyle()
        return v

    def set_alpha(self, value, a=None):
        self.setvar('alpha', value)
        fc = self._artists[0].get_facecolor()
        fc = (fc[0], fc[1], fc[2], value)
        for a in self._artists:
            a.set_alpha(value)
            a.set_facecolor(tuple(fc))

    def get_alpha(self, a):
        return self.getvar('alpha')

    def set_color(self, value, a=None):
        if self.getvar('color') == value:
            return
        for a in self._artists:
            a.set_facecolor(value)
        alpha = self.getvar('alpha')
        self.set_alpha(alpha)

    def get_color(self, a):
        return self.getvar('color')

    def get_xrange(self, xrange=[None, None], scale='linear'):
        x = self.getvar('x')
        xrange = self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])
        return xrange
        x, y, z = self._eval_xyz()
        if x is None:
            return xrange
        if scale == 'log':
            x = mask_negative(x)
        return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])

    def get_yrange(self, yrange=[None, None],
                   xrange=[None, None], scale='linear'):
        if self._n is not None:
            return self._update_range(yrange, (0, np.nanmax(self._n)))
        else:
            return yrange
#        de = self.get_data_extent()
        x, y, z = self._eval_xyz()
        if x is None:
            return yrange
        if y is None:
            return yrange
        if scale == 'log':
            y = mask_negative(y)
        return self._update_range(yrange, (np.nanmin(y), np.nanmax(y)))

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        if self.get_figaxes().get_3d():
            return zrange
        else:
            return zrange

    def _eval_xyz(self):
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return None, None, None
        return self.getp(("x", "y", "z"))

    def _saveload_names(self):
        return ['x', 'y', 'z']

    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        names = self._saveload_names()
        var = {name: check(self, name) for name in names}
        for name in names:
            if not var[name]:
                var[name+'data'] = self.getp(name)
        data['FigHist'] = (1, var)
        data = super(FigHist, self).save_data2(data)

        return data

    def load_data2(self, data):
        d = data['FigHist']
        super(FigHist, self).load_data2(data)
        var = d[1]
        names = self._saveload_names()
        for name in names:
            if not name in var:
                var[name] = True
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def get_export_val(self, a):
        return {"n":    self._n,
                "bins": self._bins}



from __future__ import print_function
#
#    merge_sol script.
#
#    it does move all children in work to a new sol,
#    which is named like case1, case2....
#
#
#    this script will be used when model does not have
#    a particular finishg_script.
#
#
#
#    copy this to your model and set finish_script in setting panel
#                          or
#    edit this file and leave the finish_script blank
#
#    this script will be called using RunA(k, worker, sol)
#    writing the data is critical part and user needs to
#    use a lock. Use PySol.aquire_lock() PySol.release_lock()
#
from ifigure.mto.py_code import PySol
sol = args[2]
worker = args[1]
index = args[0]

# sol.aquire_lock()
print(('merging solution index=', index))
print(('worker', worker))
print(('solution', sol))

sol

ps = PySol()
sol.add_child('case' + str(i), ps)

for name, child in worker.get_children():
    child.move(ps, keep_zorder=False)

# sol.release_lock()



'''
     routines for range adjustment

     purpose: create a array of actions when axes range was
              changed through GUI

              originally range adjustment action was done by
              parasit-style programming, which is difficult
              to grasp which part of program is called.
  
              idea is to orgnized the event handling in following
              steps.

                1) diagnose what event was happend and
                   construct action array directly induced
                   by the action

                   case 1 : pan: read range data from artist
                   case 2 : zoom: compute zoom from mouse event
                   case 3 : same x,y,,,
                   case 4 : auto x,y,,,
 
                   Diagnosis is done in event handler.

                   Actions are constructed by FigAxes:get_axrange_update_action,
                   which returns an array of action, and where the
                   next indirect action shold happen


                2) construct action array triggerd by the
                   direct action
                   
                   case1 : range change in colorbar should
                           change c-axes of main axes
               
                   This is done by FigAxes:get_axrange_update_action_i


                3) in multipage mode, do 1 and 2 as if the 
                   event happens in all pages.

                4) set it to history

              
    improvements
        re-computing range which was previously performed 
        everytime undo/redo is done is eliminated
 
        implementation of finish action in undo_redo_history
        is simplefied.
        
        

    keep methods:
        fig_axes.call_handle_axes_change:
        fig_obj::handle_axes_change
             a method which should be called after range parameters
             are all changed
             complication is FigColorbar is a child of main axes
        fig_axes.adjust_axes_range

    note:
        fig_axes can have two artists (more), multiple color bars
        axis_param is the place to store the data....

'''


import weakref
import numpy as np
# request builder


class RangeRequestMaker(object):
    '''
    routines for canvas
    '''
    @staticmethod
    def make_range_request_pan(fig_ax, auto=True, requests=None):
        #
        #  called from pan mode
        #
        #  fig_ax.call_handle_axes_change needs to be called
        #  after performing this request
        #
        actions = []
        arr = fig_ax._xaxis+fig_ax._yaxis
        if fig_ax._3D:
            arr.extend(fig_ax._zaxis)

        if requests is None:
            requests = {}
        if not fig_ax in requests:
            requests[fig_ax] = []

        for x in arr:
            for a in fig_ax.get_axes_artist_by_name(x.name):
                value = x.get_rangeparam()
                value[1] = auto
                value[2] = x.get_mpl_range(a)
                value[6] = False
                value[7] = False
                value[8] = False
                requests[fig_ax].append((x.name, value))
        return requests

    @staticmethod
    def make_range_request_zoom(fig_ax, direction, range, auto, ax, requests=None):
        #
        #  called from zoom mode
        #
        #
        #  fig_ax.call_handle_axes_change needs to be called
        #  after performing this request
        #

        if direction == 'x':
            arr = fig_ax._xaxis
        elif direction == 'y':
            arr = fig_ax._yaxis
        elif direction == 'z':
            arr = fig_ax._zaxis
        if requests is None:
            requests = {}

        if not fig_ax in requests:
            requests[fig_ax] = []
        for x in arr:
            if not ax in fig_ax.get_axes_artist_by_name(x.name):
                continue
            value = x.get_rangeparam()
            value[1] = auto
            value[2] = range
            value[6] = False
            value[7] = False
            value[8] = False
            requests[fig_ax].append((x.name, value))
        return requests

    @staticmethod
    def make_xyauto_request(fig_ax, direction, requests=None, scale='default'):
        # direction should be name[0]
        # if auto : call adjust_axes_range after this
        # scale default : no change

        attr = '_'+direction+'axis'
        g = getattr(fig_ax, attr)
        if requests is None:
            requests = {}
        if not fig_ax in requests:
            requests[fig_ax] = []
        for x in g:
            value = x.get_rangeparam()
            value[1] = True
            if scale != 'default':
                value[3] = scale
            requests[fig_ax].append((x.name, value))
        return requests

    @staticmethod
    def make_samexy_request(fig_ax, direction, auto, mode=1, requests=None):
        # direction should be name[0]
        # if auto : call adjust_axes_range after this
        #    not auto: call call_handle_axes_change after this
        # this should be called for each axes.
        attr = '_'+direction+'axis'
        if not hasattr(fig_ax, attr):
            return {}
        g = getattr(fig_ax, attr)

        ranges = [x.get_rangeparam()[2] for x in g]

        f_page = fig_ax.get_parent()
        if requests is None:
            requests = {}
        for f_ax in f_page.walk_axes():
            # if f_ax is fig_ax and mode == 1: continue
            if not hasattr(f_ax, attr):
                continue
            if not f_ax in requests:
                requests[f_ax] = []
            axparams = getattr(f_ax, attr)
            for i, g in enumerate(axparams):
                p = g.get_rangeparam()
                if i < len(ranges):
                    r = ranges[i]
                else:
                    r = ranges[-1]
                p[1] = auto
                if mode == 1:
                    p[2] = r
                requests[f_ax].append((g.name, p))
        # print requests
        if auto:
            for key in requests:
                requests[key] = key.compute_new_range(request=requests[key])
        return requests

    def make_autox_autoy_request(self, fig_ax):
        f_page = fig_ax.get_parent()
        requests = {}
        for f_ax in f_page.walk_axes():
            # if f_ax is fig_ax and mode == 1: continue
            requests[f_ax] = []
            axparams = getattr(f_ax, '_xaxis')
            for i, g in enumerate(axparams):
                p = g.get_rangeparam()
                p[1] = True
                requests[f_ax].append((g.name, p))
            ayparams = getattr(f_ax, '_yaxis')
            for i, g in enumerate(ayparams):
                p = g.get_rangeparam()
                p[1] = True
                requests[f_ax].append((g.name, p))

        for key in requests:
            requests[key] = key.compute_new_range(request=requests[key])
        return requests

    def make_samex_autoy_request(self, fig_ax):
        # direction should be name[0]
        # if auto : call adjust_axes_range after this
        #    not auto: call call_handle_axes_change after this
        # this should be called for each axes.
        g = getattr(fig_ax, '_xaxis')
        ranges = [x.get_rangeparam()[2] for x in g]

        f_page = fig_ax.get_parent()
        requests = {}
        for f_ax in f_page.walk_axes():
            # if f_ax is fig_ax and mode == 1: continue
            requests[f_ax] = []
            axparams = getattr(f_ax, '_xaxis')
            for i, g in enumerate(axparams):
                p = g.get_rangeparam()
                if i < len(ranges):
                    r = ranges[i]
                else:
                    r = ranges[-1]
                p[1] = False
                p[2] = r
                requests[f_ax].append((g.name, p))
            ayparams = getattr(f_ax, '_yaxis')
            for i, g in enumerate(ayparams):
                p = g.get_rangeparam()
                p[1] = True
                requests[f_ax].append((g.name, p))

        # print requests
        for key in requests:
            requests[key] = key.compute_new_range(request=requests[key])
        return requests

    @staticmethod
    def _add_induced_range_request(requests):
        irequests = {}
        for key in requests:
            d = key.get_induced_range_request(requests[key])
            for key2 in d:
                irequests[key2] = d[key2]
        for key in irequests:
            if not key in requests:
                requests[key] = []
            requests[key].extend(irequests[key])
        return requests

    def _create_multipage_rangerequest(self, requests):
        page = self._figure.figobj
        book = self._figure.figobj.get_figbook()

        mrequests = {}
        for p in book.walk_page():
            #            print requests
            for key in requests:
                iax = page.get_iaxes(key)
                ax = p.get_axes(iax)
                if ax is None:
                    continue

                mrequests[ax] = requests[key]
        return mrequests

    def expand_requests(self, requests):
        window = self.GetTopLevelParent()
        if window._use_samerange:
            mrequests = self._create_multipage_rangerequest(requests)

        requests = self._add_induced_range_request(requests)
        if window._use_samerange:
            mrequests = self._add_induced_range_request(mrequests)
            for key in mrequests:
                requests[key] = mrequests[key]
        return requests

    def send_range_action(self, requests, menu_name='edit', extra_actions=None):
        from ifigure.widgets.undo_redo_history import GlobalHistory

        a = []
        f = []
        for key in requests:
            if len(requests[key]) > 0:
                a.extend(key.make_range_actions(requests[key]))
                name = requests[key][0][0]
                f.append((weakref.ref(key), 'call_handle_axes_change', (name,)))
#            f.append((ifigure.events, 'SendRangeChangedEvent', (key,)))
        if len(a) == 0:
            return
        if extra_actions is not None:
            a.extend(extra_actions)
        window = self.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(
            a, finish_action=f, menu_name=menu_name)


class AdjustableRangeHolder(object):
    '''
    a class to be inheriged by FigAxes
    '''

    def make_range_actions(self, request):
        from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod

        actions = []
        for x in request:
            p = self.get_axis_param(x[0])
            if not self.isempty():
                if p is None:
                    continue
                a = self.get_axes_artist_by_name(x[0])[0]
                actions.append(p.make_rangeparam_action(a,  x[1]))
            else:
                action = UndoRedoFigobjMethod(None,
                                              'axrangeparam', x[1], figobj=self)
                action.set_extrainfo(x[0])
                actions.append(action)
#        x = request[-1]
#        p = self.get_axis_param(x[0])
#        a = self.get_axes_artist_by_name(x[0])[0]
#        actions.append(p.make_rangeparam_action0(a,  x[1]))
        return actions

    def perform_range_request(self, request):
        if len(request) == 0:
            return
        for x in request:
            self.set_axrangeparam(x)

        # if artist is not realized, do not call handle_axes_change
        if self.isempty():
            return
        for child in self.walk_tree():
            if child is not self:
                child.handle_axes_change({'td': self, 'name': x[0]})

#        self.set_axrangeparam0(request[-1][0])

    def get_induced_range_request(self, request):
        return {}

    def adjust_axes_range(self, get_action=False):
        use_auto = True
        try:
            if self.get_figbook()._lock_scale:
                use_auto = False
        except:
            pass
        request = self.compute_new_range(use_auto=use_auto)

        if get_action:
            actions = self.make_range_actions(request)
            return actions
        else:
            self.perform_range_request(request)

    def compute_new_range(self, request=None, use_auto=True):
        def en_int(range0):
            import numpy as np
            a = range0[0]
            b = range0[1]
            if a != 0:
                si = a/abs(a)
                ex = int(np.log10(abs(a)))
#               if (a/(10.**ex) % 1)== 0.:
                ai = (np.floor(a/(10.**ex)))*10.**ex
#               else:
#                  ai = (np.floor(a/(10.**ex))-si)*10.**ex
            else:
                ai = 0.
            if b != 0:
                si = b/abs(b)
                ex = int(np.log10(abs(b)))
                if (b/(10.**ex) % 1) == 0.:
                    bi = (np.floor(b/(10.**ex)))*10.**ex
                else:
                    bi = (np.floor(b/(10.**ex))+1)*10.**ex
            else:
                bi = 0.
            return ai, bi

        def en_sym(range0):
            a = range0[0]
            b = range0[1]
            if abs(a) > abs(b):
                return (-abs(a), abs(a))
            else:
                return (-abs(b), abs(b))

        def en_int_sym(range, mode):
            if range[0] is not None:
                if np.iscomplex(range[0]):
                    range = sorted([float(np.real(range[0])),
                                    float(np.real(range[1]))])
                if mode[2]:
                    d = abs(float(range[0])-float(range[1]))
                    if range[1] > range[0]:
                        range[1] = range[1]+d/10.
                        range[0] = range[0]-d/10.
                    else:
                        range[1] = range[1]-d/10.
                        range[0] = range[0]+d/10.
            if (mode[0] and range[0] is not None and
                    range[1] is not None):
                range = en_int(range)
            if (mode[1] and range[0] is not None and
                    range[1] is not None):
                range = en_sym(range)
            return range

        def _value2param(value):
            return (value[0], value[1], value[2], value[3], value[4], value[5],
                    (value[6], value[7], value[8]))

        def _a2param(ax):
            return (ax.base, ax.auto, ax.range, ax.scale, ax.symloglin,
                    ax.symloglinscale, ax.mode)

        data = []
        newrange = {}

        max_range_width = 0.0
        # 0) first do xrange
        for ax in self._xaxis:
            base, auto, range, scale, symloglin, symscale, mode = _a2param(ax)
            if request is not None:
                for name, value in request:
                    if name == ax.name:
                        base, auto, range, scale, symloglin, symscale, mode = _value2param(
                            value)
                        break
            if (auto and use_auto) or range is None:
                range = [None]*2
                for m in ax.walk_member():
                    if m.is_suppress():
                        continue
                    range = m.get_xrange(range, scale=scale)
                range = en_int_sym(range, mode)
                if (range[0] is None or
                        range[1] is None):
                    range = (0, 1)
                if (range[0] == range[1]):
                    range = (range[0]-0.5, range[0]+0.5)
            p = [base, auto, range, scale, symloglin, symscale, ] + list(mode)
            newrange[ax] = range

            max_range_width = max(max_range_width, abs(range[1]-range[0]))
            data.append((ax.name, p))
        # 1) second do yrange
        for ay in self._yaxis:
            base, auto, range, scale, symloglin, symscale, mode = _a2param(ay)
            if request is not None:
                for name, value in request:
                    if name == ay.name:
                        base, auto, range, scale, symloglin, symscale, mode = _value2param(
                            value)
                        break
            if (auto and use_auto) or range[0] is None:
                range = [None]*2
                for m in ay.walk_member():
                    if m.is_suppress():
                        continue
                    ax = m.get_xaxisparam()
                    if ax in newrange:
                        xrange = newrange[ax]
                    else:
                        xrange = ax.range
                    range = m.get_yrange(range,
                                         xrange=xrange, scale=scale)
                range = en_int_sym(range, mode)
                if (range[0] is None or
                        range[1] is None):
                    range = (0, 1)
                if (range[0] == range[1]):
                    range = (range[0]-0.5, range[0]+0.5)
            p = [base, auto, range, scale, symloglin, symscale, ] + list(mode)
            newrange[ay] = range

            max_range_width = max(max_range_width, abs(range[1]-range[0]))
            data.append((ay.name, p))
        # 2-1) third do zrange
        for az in self._zaxis:
            base, auto, range, scale, symloglin, symscale, mode = _a2param(az)
            if request is not None:
                for name, value in request:
                    if name == az.name:
                        base, auto, range, scale, symloglin, symscale, mode = _value2param(
                            value)
                        break
            if (auto and use_auto) or range[0] is None:
                range = [None]*2
                for m in az.walk_member():
                    if m.is_suppress():
                        continue
                    ax = m.get_xaxisparam()
                    ay = m.get_yaxisparam()
                    if ax in newrange:
                        xrange = newrange[ax]
                    else:
                        xrange = ax.range
                    if ay in newrange:
                        yrange = newrange[ay]
                    else:
                        yrange = ay.range
                    range = m.get_zrange(range, xrange=xrange,
                                         yrange=yrange, scale=scale)
                range = en_int_sym(range, mode)
                if (range[0] is None or
                        range[1] is None):
                    range = (0, 1)
                if (range[0] == range[1]):
                    #range = (range[0]-0.5, range[0]+0.5)
                    range = (range[0]-max_range_width/2.,
                             range[0]+max_range_width/2.)
            p = [base, auto, range, scale, symloglin, symscale] + list(mode)
            data.append((az.name, p))
        # 2-2) third do crange
        for ac in self._caxis:
            base, auto, range, scale, symloglin, symscale, mode = _a2param(ac)
            if request is not None:
                for name, value in request:
                    if name == ac.name:
                        base, auto, range, scale, symloglin, symscale, mode = _value2param(
                            value)
                        break

            if (auto and use_auto) or range[0] is None:
                range = [None]*2
                for m in ac.walk_member():
                    if m.is_suppress():
                        continue
                    ax = m.get_xaxisparam()
                    ay = m.get_yaxisparam()
                    if ax in newrange:
                        xrange = newrange[ax]
                    else:
                        xrange = ax.range
                    if ay in newrange:
                        yrange = newrange[ay]
                    else:
                        yrange = ay.range
                    range = m.get_crange(range, xrange=xrange,
                                         yrange=yrange,
                                         scale=scale)
                range = en_int_sym(range, mode)
                if (range[0] is None or
                        range[1] is None):
                    range = (0, 1)
                if (range[0] == range[1]):
                    if range[0] == 0:
                        delta = 1 - np.nextafter(1.0, 0.0)
                        range = (-abs(delta), abs(delta))
                    else:
                        range = (range[0]-abs(range[0])/10,
                                 range[0]+abs(range[0])/10)
            p = [base, auto, range, scale, symloglin, symscale, ] + list(mode)
            data.append((ac.name, p))
        return data


class AdjustableRangeHolderCbar(AdjustableRangeHolder):
    '''
    routie for cbar
    '''

    def get_induced_range_request(self, request):
        '''
        range change in cbar will make additional
        request in main axes
        '''

        if self.getp("cdir") == 'v':
            self._artists[0].set_xlim((0, 1))
            for name, value in request:
                if name.startswith('y'):
                    cmin, cmax = value[2]
        else:
            self._artists[0].set_ylim((0, 1))
            for name, value in request:
                if name.startswith('x'):
                    cmin, cmax = value[2]

        # ifigure.events.SendIdleEvent(self)
        value = self._caxis_param().get_rangeparam()

#        if (value[2][0] != cmin or
#            value[2][1] != cmax):
        value[1] = False
        value[2] = [cmin, cmax]
        return {self.get_parent():
                [(self._caxis_param().name, value), ]}

#        return {}



import wx
import wx.stc as stc
import keyword
#import  ifigure.icon.images as images
from ifigure.widgets.script_editor import faces, PythonSTC
import ifigure


class HistoryPopUp(wx.Menu):
    def __init__(self, parent, reload=False):
        super(HistoryPopUp, self).__init__()
        menus = [  # ('Cut',  parent.onCut, None),
            ('Copy', parent.onCopy, None),
            ('Copy To Shell', parent.onCopyToShell, None),
            #                 ('Paste', parent.onPaste, None),
            #                 ('Delete', parent.onDeleteBack, None),
            #                 ('---', None, None,),
            ('Select All', parent.onSelectAll, None),
            ('---', None, None,),
            ('Run Selection', parent.onRunSelection, None), ]

#                 ('Export to command history', parent.onExportHistoryToCommand, None)]
        ifigure.utils.cbook.BuildPopUpMenu(self, menus, eventobj=parent)


class LinePythonSTC(PythonSTC):
    def __init__(self, *args, **kargs):
        PythonSTC.__init__(self, *args, **kargs)
        self.MarkerSetForeground(1, 'GOLD')
        self.MarkerSetBackground(1, 'GOLD')
        self._marker1s = []
        self._margins = [self.GetMarginWidth(
            0), self.GetMarginWidth(1), self.GetMarginWidth(2)]
        self._get_down = -1
        self._line_down = -1
        self._marked_line = (-1, -1)
        self._dnd_started = False
        self.Bind(wx.EVT_LEFT_DOWN, self.onLeftDown)
        self.Bind(wx.EVT_LEFT_UP, self.onLeftUp)
        self.Bind(wx.EVT_RIGHT_UP, self.onRightUp)

    def x2margin(self, x):
        if x < self._margins[0]:
            return 0
        elif (x >= self._margins[0] and
              x < self._margins[0] + self._margins[1]):
            return 1
        elif (x >= self._margins[0] + self._margins[1] and
              x < self._margins[0] + self._margins[1] + self._margins[2]):
            return 2
        else:
            return 3

    def OnKeyPressed(self, event):
        def SE_Cut():
            'cut for script editor'
            self.GetTopLevelParent().SetStatusText('cut')
            if self.GetSelectedText() == '':
                pos2 = self.GetCurrentPos()
                self.SetSelection(self._mark, pos2)
            self.Cut()
            self._ctrl_K = False
            return

        def SE_Copy():
            self.GetTopLevelParent().SetStatusText('copy')
            sel = self.GetSelection()
            if self.GetSelectedText() == '':
                pos2 = self.GetCurrentPos()
                self.SetSelection(self._mark, pos2)
            self.Copy()
            self.SetSelection(sel[0], sel[1])
            self._ctrl_K = False
            return

        if self.CallTipActive():
            self.CallTipCancel()
        if self.AutoCompActive():
            self.AutoCompCancel()

        key = event.GetKeyCode()
        if hasattr(event, 'RawControlDown'):
            controlDown = event.RawControlDown()
        else:
            controlDown = event.ControlDown()
        altDown = event.AltDown()

        if (self._search != 0 and
            not controlDown and
                not altDown):

            event.Skip()
            return
        if (self._search != 0 and
                not (key == 71 and controlDown)):   # not  control G
            # in search mode ctrl+S or ctrl+R are
            # the only option otherwise
            # it exits from search mode
            # ctrl+G in search mode is treated later...
            if key == wx.WXK_SHIFT:
                event.Skip()
                return
            if key == wx.WXK_CONTROL:
                event.Skip()
                return
            if key == wx.WXK_RAW_CONTROL:
                event.Skip()
                return
            if not ((key == 83 and controlDown) or
                    (key == 82 and controlDown)):
                self._search = 0
                self._search_st = -1
                self.set_search_text('')

        if key == wx.WXK_SPACE:
            self._mark = self.GetCurrentPos()
            self.GetTopLevelParent().SetStatusText('mark set')

        if key == 32:
            pass
        elif key == 68 and controlDown:  # ctrl + D (delete)
            self.CharRight()
            self.DeleteBackNotLine()
            return

        elif key == 67 and controlDown:  # ctrl + C (copy)
            SE_Copy()
            return

        elif key == 86 and controlDown:  # ctrl + V ( skip 10 lines)
            tmp = self.GetCurrentLine()
            c = 0
            while tmp < self.GetLineCount() and c < 10:
                #               print tmp, self.GetLineVisible(tmp)
                tmp = tmp+1
                if self.GetLineVisible(tmp):
                    c = c+1

            self.GotoLine(tmp)

            return

        elif key == 87 and event.AltDown():  # alt + W (copy)
            SE_Copy()
            return

        elif key == 69 and controlDown:  # ctrl + E (end of line)
            self.LineEnd()
            return
        elif key == 78 and controlDown:  # control + N (next command)
            self.LineDown()
            return
        elif key == 80 and controlDown:  # ctrl + P (prev command)
            self.LineUp()
            return
        elif key == 65 and controlDown:  # ctrl + A (beginning)
            self.Home()
            return
        elif key == 70 and controlDown:  # ctrl + F (forward)
            self.CharRight()
            return
        elif key == 66 and controlDown:  # ctrol + B (back)
            self.CharLeft()
            return
        elif key == 83 and controlDown:  # ctrl + S (search)
            #            print('forward search', self._search)

            if self._search == 0:
                self._search = 1
                self.set_search_text('')
                self._search_st = self.GetCurrentPos()
            else:
                self._search = 1

            tmp = self.GetCurrentPos()
            for i in range(len(self._search_txt)):
                self.CharRight()
            self.SearchAnchor()

            if self._search_txt != '':
                if self.SearchNext(0, self._search_txt) == -1:
                    self.set_search_status_text(fail=True)
                    self.SetCurrentPos(tmp)
                    self.SetSelectionEnd(tmp)
                    self.SetSelectionStart(tmp)
            self.EnsureCaretVisible()
#            self.Bind(wx.EVT_CHAR, self.onSearch)
            return

        elif key == 82 and controlDown:  # ctrl + R (search)
            if self._search == 0:
                self._search = 2
                self.set_search_text('')
                self._search_st = self.GetCurrentPos()
            else:
                self._search = 2
            self.SearchAnchor()
            if self._search_txt != '':
                if self.SearchPrev(0, self._search_txt) == -1:
                    self.set_search_status_text(fail=True)
            self.EnsureCaretVisible()
#            self.Bind(wx.EVT_CHAR, self.onSearch)
            return

        elif key == 71 and controlDown:  # ctrl + G (exit search)
            self._mark = -1
            self._search = 0
            self.set_search_text('')
#            self.SetCurrentLine(self._search_st[0])
            if self._search_st != -1:
                self.SetCurrentPos(self._search_st)
                self.SetSelectionEnd(self._search_st)
                self.SetSelectionStart(self._search_st)
                self.EnsureCaretVisible()
            self._search_st = -1
#            self.Unbind(wx.EVT_CHAR)
            return
        event.Skip()

    def update_marker(self, l1, l2):
        if len(self._marker1s) != 0:
            self.MarkerDeleteAll(1)
            self._marker1s = []
            if l1 == l2:
                return
        for x in range(l2-l1+1):
            self._marker1s.append(self.MarkerAdd(l1+x, 1))
        self._marked_line = (l1, l2)

    def reset_selection(self):
        sels, sele = self.GetSelection()
        self.SetSelection(sels, sele)
        self.Refresh()

    def onLeftDown(self, e):
        self._line_down = self.LineFromPosition(
            self.PositionFromPoint(e.GetPosition()))
        self._get_down = self.PositionFromPoint(e.GetPosition())
        e.Skip()

    def onLeftUp(self, e):
        #        print 'left up'
        #        print self.PositionFromPoint(e.GetPosition())
        if self._get_down == self.PositionFromPoint(e.GetPosition()):
            st, et = self.GetSelection()
#            print 'here', st, et
#            print self.LineFromPosition(st)
#            print self.LineFromPosition(self.PositionFromPoint(e.GetPosition()))
            if (st != et and
                self.LineFromPosition(st) ==
                    self.LineFromPosition(self.PositionFromPoint(e.GetPosition()))):
                wx.CallAfter(self.SetSelection, self._get_down, self._get_down)
#               wx.CallAfter(self.update_marker, 0,0)
#               self.SetSelection(0,0)
                e.Skip()
                return
        l1 = self.LineFromPosition(self._get_down)
        l2 = self.LineFromPosition(self.PositionFromPoint(e.GetPosition()))
#        print(l1, l2)
        ls = min([l1, l2])
        le = max([l1, l2])
        if ls == 0:
            st = 0
        else:
            st = self.GetLineEndPosition(ls-1) + 1
        et = self.GetLineEndPosition(le)

        wx.CallAfter(self.SetSelection, st, et)
#        wx.CallAfter(self.update_marker, l1, l2)
        e.Skip()

    def onDragInit(self, e):
        self.Unbind(wx.EVT_MOTION)

    def OnStartDrag(self, evt):
        sel = self.GetSelectedText()
        sel = '\n'.join([x[1:]
                         for x in sel.split('\n') if not x.startswith("'''")])
        # evt.SetDragAllowMove(False)
        evt.SetDragText(sel)
        p = self
        while p.GetParent() is not None:
            p = p.GetParent()

        p._text_clip = sel

    def onRightUp(self, evt):
        m = HistoryPopUp(self)
        self.PopupMenu(m,
                       (evt.GetX(), evt.GetY()))
        m.Destroy()

    def onExportHistoryToCommand(self):
        pass

    def onCopyToShell(self, evt):
        '''
        need to over write to remove leading space 
        '''
        sp, ep = self.GetSelection()
        l1 = self.LineFromPosition(sp)
        l2 = self.LineFromPosition(ep)
        line = []
        eol = self.get_EOL_txt()
        if l1 == l2:
            line = self.GetTextRange(sp, ep)
            if len(line) == 0:
                return
            line = [line]
        else:
            for i in range(l2-l1+1):
                line.append(str(self.GetLine(l1+i)).strip(eol))
        if len(line) == 0:
            return
        line = [x[1:] for x in line if not x.startswith("'''")]

        app = wx.GetApp().TopWindow  # self.GetTopLevelParent()
        for i, x in enumerate(line):
            self.CopyText(len(x), x)
            app.shell.Paste()
            if i != len(line)-1:
                app.shell.NewLine()


class CommandHistory(wx.Panel):
    def __init__(self, parent, *args, **kargs):
        super(CommandHistory, self).__init__(parent, *args, **kargs)
        self.log = LinePythonSTC(
            self, -1, style=wx.TE_MULTILINE | wx.TE_READONLY)

        self.log.SetSelectionMode(wx.stc.STC_SEL_LINES)

        self.log.SetReadOnly(True)
#        sys.stdout = RedirectOutput(self.log)
#        sys.stderr = RedirectOutput(self.log)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.log, 1, wx.EXPAND)
        self.SetSizer(sizer)
        self.clear_text()

#        self.log.Bind(wx.EVT_LEFT_DOWN, self.onLeftDown)
#        self.log.Bind(wx.EVT_LEFT_UP, self.onLeftUp)

    def Copy(self):
        self.log.Copy()

    def Cut(self):
        # map cut to copy
        self.log.Copy()

    def Paste(self):
        pass
        # self.log.Copy()

    def date_string(self):
        import datetime
        return "''' " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M") + "\n'''\n"

    def savetofile(self, filename=''):
        fid = open(filename, 'w')
        txt = self.log.GetText()
        fid.write(txt)
        fid.close()

    def loadfromfile(self, filename=''):
        fid = open(filename, 'r')
        lines = fid.readlines()
        fid.close()
        self.log.SetReadOnly(False)
        self.log.SetText(''.join(lines))
        self.log.SetReadOnly(True)
        self.append_text(self.date_string(), no_indent=True)

    def clear_text(self):
        self.log.SetReadOnly(False)
        self.log.SetText(self.date_string())
        self.log.SetReadOnly(True)

    def append_text(self, new_txt, no_indent=False):
        self.log.SetReadOnly(False)
        txt = self.log.GetText()
        if (not txt.endswith('\n') and
                txt[-2:] != '\n\n'):
            txt += '\n'
        if no_indent:
            txt += str(new_txt)
        else:
            for x in new_txt.split('\n'):
                txt += ' '+str(x) + '\n'
            txt = txt[:-1]
        l = self.log.GetFirstVisibleLine()
        self.log.SetText(txt)
        self.log.SetReadOnly(True)
        self.log.ScrollToLine(l)



# A part of pdfrw (pdfrw.googlecode.com)
# Copyright (C) 2006-2012 Patrick Maupin, Austin, Texas
# MIT license -- See LICENSE.txt for details

'''

This module contains code to build PDF "Form XObjects".

A Form XObject allows a fragment from one PDF file to be cleanly
included in another PDF file.

Reference for syntax: "Parameters for opening PDF files" from SDK 8.1

        http://www.adobe.com/devnet/acrobat/pdfs/pdf_open_parameters.pdf

        supported 'page=xxx', 'viewrect=<left>,<top>,<width>,<height>'

        Also supported by this, but not by Adobe:
            'rotate=xxx'  where xxx in [0, 90, 180, 270]

        Units are in points


Reference for content:   Adobe PDF reference, sixth edition, version 1.7

        http://www.adobe.com/devnet/acrobat/pdfs/pdf_reference_1-7.pdf

        Form xobjects discussed chapter 4.9, page 355
'''

from pdfrw.objects import PdfDict, PdfArray, PdfName
from pdfrw.pdfreader import PdfReader
from pdfrw.errors import log
from six import iteritems

class ViewInfo(object):
    ''' Instantiate ViewInfo with a uri, and it will parse out
        the filename, page, and viewrect into object attributes.
    '''
    doc = None
    docname = None
    page = None
    viewrect = None
    rotate = None

    def __init__(self, pageinfo='', **kw):
        pageinfo = pageinfo.split('#', 1)
        if len(pageinfo) == 2:
            pageinfo[1:] = pageinfo[1].replace('&', '#').split('#')
        for key in 'page viewrect'.split():
            if pageinfo[0].startswith(key+'='):
                break
        else:
            self.docname = pageinfo.pop(0)
        for item in pageinfo:
            key, value = item.split('=')
            key = key.strip()
            value = value.replace(',', ' ').split()
            if key in ('page', 'rotate'):
                assert len(value) == 1
                setattr(self, key, int(value[0]))
            elif key == 'viewrect':
                assert len(value) == 4
                setattr(self, key, [float(x) for x in value])
            else:
                log.error('Unknown option: %s', key)
        for key, value in iteritems(kw):
            assert hasattr(self, key), key
            setattr(self, key, value)


def get_rotation(rotate):
    ''' Return clockwise rotation code:
          0 = unrotated
          1 = 90 degrees
          2 = 180 degrees
          3 = 270 degrees
    '''
    try:
        rotate = int(rotate)
    except (ValueError, TypeError):
        return 0
    if rotate % 90 != 0:
        return 0
    return rotate / 90


def rotate_point(point, rotation):
    ''' Rotate an (x,y) coordinate clockwise by a 
        rotation code specifying a multiple of 90 degrees.
    '''
    if rotation & 1:
        point = point[1], -point[0]
    if rotation & 2:
        point = -point[0], -point[1]
    return point


def rotate_rect(rect, rotation):
    ''' Rotate both points within the rectangle, then normalize
        the rectangle by returning the new lower left, then new
        upper right.
    '''
    rect = rotate_point(rect[:2], rotation) + rotate_point(rect[2:], rotation)
    return (min(rect[0], rect[2]), min(rect[1], rect[3]),
            max(rect[0], rect[2]), max(rect[1], rect[3]))


def getrects(inheritable, pageinfo, rotation):
    ''' Given the inheritable attributes of a page and
        the desired pageinfo rectangle, return the page's
        media box and the calculated boundary (clip) box.
    '''
    mbox = tuple([float(x) for x in inheritable.MediaBox])
    vrect = pageinfo.viewrect
    if vrect is None:
        cbox = tuple([float(x) for x in (inheritable.CropBox or mbox)])
    else:
        # Rotate the media box to match what the user sees,
        # figure out the clipping box, then rotate back
        mleft, mbot, mright, mtop = rotate_rect(mbox, rotation)
        x, y, w, h = vrect
        cleft = mleft + x
        ctop = mtop - y
        cright = cleft + w
        cbot = ctop - h
        cbox = max(mleft, cleft), max(mbot, cbot), min(
            mright, cright), min(mtop, ctop)
        cbox = rotate_rect(cbox, -rotation)
    return mbox, cbox


def _cache_xobj(contents, resources, mbox, bbox, rotation):
    ''' Return a cached Form XObject, or create a new one and cache it.
        Adds private members x, y, w, h
    '''
    cachedict = contents.xobj_cachedict
    if cachedict is None:
        cachedict = contents.private.xobj_cachedict = {}
    cachekey = mbox, bbox, rotation
    result = cachedict.get(cachekey)
    if result is None:
        func = (_get_fullpage, _get_subpage)[mbox != bbox]
        result = PdfDict(
            func(contents, resources, mbox, bbox, rotation),
            Type=PdfName.XObject,
            Subtype=PdfName.Form,
            FormType=1,
            BBox=PdfArray(bbox),
        )
        rect = bbox
        if rotation:
            matrix = rotate_point((1, 0), rotation) + \
                rotate_point((0, 1), rotation)
            result.Matrix = PdfArray(matrix + (0, 0))
            rect = rotate_rect(rect, rotation)

        result.private.x = rect[0]
        result.private.y = rect[1]
        result.private.w = rect[2] - rect[0]
        result.private.h = rect[3] - rect[1]
        cachedict[cachekey] = result
    return result


def _get_fullpage(contents, resources, mbox, bbox, rotation):
    ''' fullpage is easy.  Just copy the contents,
        set up the resources, and let _cache_xobj handle the
        rest.
    '''
    return PdfDict(contents, Resources=resources)


def _get_subpage(contents, resources, mbox, bbox, rotation):
    ''' subpages *could* be as easy as full pages, but we
        choose to complicate life by creating a Form XObject
        for the page, and then one that references it for
        the subpage, on the off-chance that we want multiple
        items from the page.
    '''
    return PdfDict(
        stream='/FullPage Do\n',
        Resources=PdfDict(
            XObject=PdfDict(
                FullPage=_cache_xobj(contents, resources, mbox, mbox, 0)
            )
        )
    )


def pagexobj(page, viewinfo=ViewInfo(), allow_compressed=True):
    ''' pagexobj creates and returns a Form XObject for
        a given view within a page (Defaults to entire page.)
    '''
    inheritable = page.inheritable
    resources = inheritable.Resources
    rotation = get_rotation(inheritable.Rotate)
    mbox, bbox = getrects(inheritable, viewinfo, rotation)
    rotation += get_rotation(viewinfo.rotate)
    contents = page.Contents
    # Make sure the only attribute is length
    # All the filters must have been executed
    assert int(contents.Length) == len(contents.stream)
    if not allow_compressed:
        assert len([x for x in contents.items()]) == 1
    return _cache_xobj(contents, resources, mbox, bbox, rotation)


def docxobj(pageinfo, doc=None, allow_compressed=True):
    ''' docxobj creates and returns an actual Form XObject.
        Can work standalone, or in conjunction with
        the CacheXObj class (below).
    '''
    if not isinstance(pageinfo, ViewInfo):
        pageinfo = ViewInfo(pageinfo)

    # If we're explicitly passed a document,
    # make sure we don't have one implicitly as well.
    # If no implicit or explicit doc, then read one in
    # from the filename.
    if doc is not None:
        assert pageinfo.doc is None
        pageinfo.doc = doc
    elif pageinfo.doc is not None:
        doc = pageinfo.doc
    else:
        doc = pageinfo.doc = PdfReader(
            pageinfo.docname, decompress=not allow_compressed)
    assert isinstance(doc, PdfReader)

    sourcepage = doc.pages[(pageinfo.page or 1) - 1]
    return pagexobj(sourcepage, pageinfo, allow_compressed)


class CacheXObj(object):
    ''' Use to keep from reparsing files over and over,
        and to keep from making the output too much
        bigger than it ought to be by replicating
        unnecessary object copies.
    '''

    def __init__(self, decompress=False):
        ''' Set decompress true if you need
            the Form XObjects to be decompressed.
            Will decompress what it can and scream
            about the rest.
        '''
        self.cached_pdfs = {}
        self.decompress = decompress

    def load(self, sourcename):
        ''' Load a Form XObject from a uri
        '''
        info = ViewInfo(sourcename)
        fname = info.docname
        pcache = self.cached_pdfs
        doc = pcache.get(fname)
        if doc is None:
            doc = pcache[fname] = PdfReader(fname, decompress=self.decompress)
        return docxobj(info, doc, allow_compressed=not self.decompress)



from __future__ import print_function
#
#   Threaded worker
#      A utility class to support thread-based parallelism
#      in model execution.
#

import time
import ifigure
import wx
from six.moves import queue as Queue
import threading
from ifigure.utils.event_driven_thread2 import get_thread, send_event


def c_thread():
    import threading
    return threading.current_thread()


class ThreadedWorker(object):
    def __init__(self):
        self._waitlist1 = {}
        self._waitlist2 = {}
        self._background = False
        self._run_after = ''
        self._restart = False
        self._run_mode = False
        self._run_verbose = False
        self._return_queue = None

    def get_background(self):
        return self._background

    def set_background(self, value=True):
        self._background = value
        return self

    def _initialize_waitlist(self):
        self._waitlist1 = {}
        self._waitlist2 = {}
        self._restart = False

        for c in self.walk_model():
            c._initialize_waitlist()

    def Run(self, event=None, wait=None, return_queue=None):
        '''
        setup thread
        event combination
        '''
        self._initialize_waitlist()

        if wait is not None:
            return_queue = Queue.Queue()
        t, queue = get_thread(return_queue)
        t._verbose = self._run_verbose
        trigger = t.bind_init(self, self._doJob1)
        t_list = self.set_background(False)._setup_jobchain(t, [t])

        if event is not None:
            w = event.GetEventObject()
        else:
            w = self.get_app()
        for t in t_list:
            ifigure.events.SendThreadStartEvent(self, w=w, thread=t)
        # print 'thread list', t_list

        send_event(trigger)

        if wait is not None:
            t = threading.current_thread()
            if t.name == 'MainThread':
                while return_queue.empty():
                    wx.Yield()
            else:
                return_queue.get()
        if return_queue is not None:
            return [t.name for t in t_list]
#            if hasattr(wait, 'put'): wait.put(self)

    def _setup_jobchain(self, current_thread, thread_list):
        from ifigure.mto.py_code import PyModel

        for name in self._run_after.split(','):
            if isinstance(self._parent, PyModel):
                work = self._parent.eval_setting_str(name.strip())
                if work is not None:
                    self._add_waittostart(work)

        name = self.get_full_path()
        if self._background:
            t, queue = get_thread()
            t._verbose = self._run_verbose
            t.bind(name+'.job1_done', self._doJob2)
            self._bind_waitlist1(t)
            thread_list.append(t)
            if not self._run_mode:
                for c in self.walk_model():
                    if c.is_suppress():
                        continue
                    c._add_waittostart(self, 'job1_done')
                    thread_list = c._setup_jobchain(t, thread_list)
                    self._add_waittofinish(c)
            self._bind_waitlist2(t)
        else:
            current_thread.bind(name+'.job1_done', self._doJob2)
            self._bind_waitlist1(current_thread)
            if not self._run_mode:
                for c in self.walk_model():
                    if c.is_suppress():
                        continue
                    c._add_waittostart(self, 'job1_done')
                    thread_list = c._setup_jobchain(
                        current_thread, thread_list)
                    self._add_waittofinish(c)
            self._bind_waitlist2(current_thread)
        return thread_list

    def _bind_waitlist2(self, t):
        for e in self._waitlist2:
            t.bind(e, self._doJob2)

    def _bind_waitlist1(self, t):
        for e in self._waitlist1:
            t.bind(e, self._doJob1)

    def _add_waittostart(self, work, m='job2_done'):
        name = work.get_full_path()
        self._waitlist1[name + '.' + m] = False

    def _add_waittofinish(self, work):
        name = work.get_full_path()
        self._waitlist2[name + '.job2_done'] = False

    def _doJob1(self, event):
        # check conditions
        if event in self._waitlist1:
            self._waitlist1[event] = True
        v = True
        for key in self._waitlist1:
            v = v and self._waitlist1[key]
        if not v:
            return False

        name = self.get_full_path()
        if self._run_verbose:
            print('performing ' + name + '.job1 in ' + c_thread().name)
        self._status = 'running...'
        app = wx.GetApp().TopWindow
#        print 'in do job1', self, threading.current_thread().name
        wx.CallAfter(app.proj_tree_viewer.update_widget_request2)

        self.do_run()
        send_event(name + '.job1_done')

#        for key in self._waitlist1: self._waitlist1[key] = False
        return True

    def _doJob2(self, event):
        if event in self._waitlist2:
            self._waitlist2[event] = True
        v = True
        for key in self._waitlist2:
            v = v and self._waitlist2[key]

        if not v:
            return False

        name = self.get_full_path()
        if self._run_verbose:
            print('performing ' + name + '.job2 in ' + c_thread().name)

        send_event(name + '.job2_done')

        self._status = ''
#        print 'in do job2', self, threading.current_thread().name
#        if self._return_queue is not None: self._return_queue.put('done')
        app = wx.GetApp().TopWindow
        wx.CallAfter(app.proj_tree_viewer.update_widget_request2)

#        for key in self._waitlist2: self._waitlist2[key] = False
        return True



from __future__ import print_function
import subprocess as sp
import os
import Pyro4


class ProxyNeTe(object):
    def fill_array(self, i, txt):
        ct = []
        print((i, txt[i]))
        for t in range(1+i, int(txt[i])+1+i):
            ct.append(txt[t])
        return t+1, ct

    def run_idl(self, shot=1120809022,
                time=1):

        p = sp.Popen('idl', stdin=sp.PIPE, shell=True,
                     stdout=sp.PIPE)

        p.stdin.write(
            '.r /home/shiraiwa/PycharmProjects2M/ifigure/rpo/nete_ispline.pro\n')
        p.stdin.write('v=nete_ispline('+str(shot)+','+str(time)+')\n')
        # ne(core)
        p.stdin.write('print, n_elements(v.cne)\n')
        p.stdin.write('print, reform(v.cne)\n')
        p.stdin.write('print, n_elements(v.cne_err)\n')
        p.stdin.write('print, reform(v.cne_err)\n')
        # te(core)
        p.stdin.write('print, n_elements(v.cte)\n')
        p.stdin.write('print, reform(v.cte)\n')
        p.stdin.write('print, n_elements(v.cte_err)\n')
        p.stdin.write('print, reform(v.cte_err)\n')
        # r(core)
        p.stdin.write('print, n_elements(v.c_rho)\n')
        p.stdin.write('print, reform(v.c_rho)\n')
        # ne(edge)
        p.stdin.write('print, n_elements(v.ene)\n')
        p.stdin.write('print, reform(v.ene)\n')
        p.stdin.write('print, n_elements(v.ene_err)\n')
        p.stdin.write('print, reform(v.ene_err)\n')
        # te(edge)
        p.stdin.write('print, n_elements(v.ete)\n')
        p.stdin.write('print, reform(v.ete)\n')
        p.stdin.write('print, n_elements(v.ete_err)\n')
        p.stdin.write('print, reform(v.ete_err)\n')
        # r(edge)
        p.stdin.write('print, n_elements(v.e_rho)\n')
        p.stdin.write('print, reform(v.e_rho)\n')

        p.stdin.write('exit\n')
        p.stdout.flush()

        txt = ''.join(p.stdout.read())
        txt = txt.replace('\n', ' ')
        txt = txt.replace('\r', ' ')
        txt = [float(x) for x in txt.split(' ') if len(x) != 0]
#        a = txt.split('\r')
#        txt=''.join(a)
#        a = txt.split('\n')
#        txt=''.join(a)
        print(txt)
        val = {}
        i = 0
        i, val["cn"] = self.fill_array(i, txt)
        i, val["cn_err"] = self.fill_array(i, txt)
        i, val["ct"] = self.fill_array(i, txt)
        i, val["ct_err"] = self.fill_array(i, txt)
        i, val["cr"] = self.fill_array(i, txt)
        i, val["en"] = self.fill_array(i, txt)
        i, val["en_err"] = self.fill_array(i, txt)
        i, tmp = self.fill_array(i, txt)
        val["et"] = [x/1000. for x in tmp]
        i, tmp = self.fill_array(i, txt)
        val["et_err"] = [x/1000. for x in tmp]
        i, val["er"] = self.fill_array(i, txt)

        print(val)
        return val


def main():
    p = ProxyNeTe()
    nameserver = Pyro4.locateNS()
    Pyro4.Daemon.serveSimple(
        {
            p: "proxy_nete"
        },
        ns=True,
        host=os.getenv('HOSTNAME'))


if __name__ == '__main__':
    # check program
    main()



from __future__ import print_function
from ifigure.utils.edit_list import TextCtrlCopyPaste
import wx
import numpy as np
GTC_ENTER = wx.NewEventType()
EVT_GTC_ENTER = wx.PyEventBinder(GTC_ENTER, 1)


class GrowableTextCtrl(TextCtrlCopyPaste):
    def __init__(self, *args, **kargs):
        #        kargs['style']=wx.TE_PROCESS_ENTER|wx.TE_MULTILINE
        kargs['style'] = wx.TE_MULTILINE
        super(GrowableTextCtrl, self).__init__(*args, **kargs)
        self.SetBackgroundColour('yellow')
        self._use_escape = False
        if len(args) > 2:
            self.setText(args[2])
        else:
            self.setText('')

    def onKeyPressed(self, evt):
        super(GrowableTextCtrl, self).onKeyPressed(evt)

        def check_length(evt, obj=self):
            try:
                ttt = obj.GetValue()
            except ValueError:
                obj.Unbind(wx.EVT_IDLE)
                return
            obj.setText(ttt)
            obj.Layout()
            obj.Unbind(wx.EVT_IDLE)
        self.Bind(wx.EVT_IDLE, check_length)

    def setText(self, txt):
        def screen_size(xx):
            pos = np.array(xx.GetPosition())
            size = np.array(xx.GetSize())
            return pos[0], pos[1], pos[0]+size[0], pos[1]+size[1]

        l = self.GetInsertionPoint()
        self.SetValue(txt)
        self.SetInsertionPoint(l)

        num = self.GetNumberOfLines()
        length = max(self.GetLineLength(x) for x in range(num))
        s = self.GetFullTextExtent('A')
        ss = np.array([self.GetFullTextExtent(self.GetLineText(x))
                       for x in range(num)])

        self.SetSize(
            (np.max(ss[:, 0])+25, np.max([np.max(ss[:, 1]), s[1]])*num+10))

        x0, y0, x1, y1 = screen_size(self)
        x0p, y0p, x1p, y1p = screen_size(self.GetParent())

        # moves control so that it stays inside the parent object
        # it takes care of running out of the parent box toward right.
        # it doesn't do anything if it is exceeding toweard lef too.
        if x1p < x1:
            dx = x1 - x1p + 1
            if x0 > dx:
                self.SetPosition((x0 - dx, y0))

    def onEnter(self, evt):
        evt2 = wx.PyCommandEvent(GTC_ENTER, wx.ID_ANY)
        evt2.SetEventObject(self)
        wx.PostEvent(self, evt2)


class test_frame(wx.Frame):
    def __init__(self, *args, **kargs):
        wx.Frame.__init__(self, *args, **kargs)
        panel = wx.Panel(self, wx.ID_ANY)
        panel.SetSize((350, 550))
        # self.SetSizer(wx.BoxSizer(wx.VERTICAL))
        #self.GetSizer().Add(panel, 1, wx.ALL, 0)
        #self.txt = TransparentTextCtrlPanel(self, wx.ID_ANY)
        self.txt = GrowableTextCtrl(panel, wx.ID_ANY, '')
        self.txt.SetPosition((120, 10))
        self.txt.SetSize((3, 3))
        self.Layout()
        self.Show()
        self.Bind(EVT_GTC_ENTER, self.onEnter)

    def onEnter(self, evt):
        print('hit enter')


if __name__ == '__main__':
    app = wx.App(False)
    frame = test_frame(None, wx.ID_ANY, "test")
    app.MainLoop()



#
#  user of cb (color zaxis)
#
import weakref
import ifigure


def get_axisparam(figaxes, axlist, figobj, func, cidx=0):
    for ax in axlist:
        if hasattr(ax, "_loaded_member"):
            if len(ax._loaded_member) == 0:
                del ax._loaded_member
                continue

            p = figaxes.get_td_path(figobj)[1:]
#           p = figaxes.get_relative_path(figobj)
            members = ax._loaded_member

            for m in members:
                if m == p:
                    #                 print 'found in loaded member'
                    func(ax)
                    ax._loaded_member = [t for t in members if t != p]
                    if len(ax._loaded_member) == 0:
                        del ax._loaded_member
                    return
    for a in axlist:

        if cidx in a._ax_idx:
            #      if len(a._member) == 0:
            func(a)
            return
    func(axlist[-1])


class XUser(object):
    def __init__(self):
        self._ax = None

    def get_xaxisparam(self):
        if self._ax is None:
            # create colorbar
            figaxes = self.get_figaxes()
            if figaxes is None:
                return None
            if len(figaxes._xaxis) == 0:
                figaxes.add_axis_param(dir='x')
            get_axisparam(figaxes, figaxes._xaxis, self,
                          self.set_ax, self._container_idx)
#           for ax in figaxes._xaxis:
#               if hasattr(ax, "_loaded_member"):
#                   members = ax["_loaded_member")]
#                   for m in members:
#                       if m == p:
#                           print 'found in loaded member'
#                           self.set_ax(ax)
#                           del members[m]
#                           return self._ax()
#           self.set_ax(figaxes._xaxis[0])
        return self._ax()

    def _ax_dead(self, ref):
        self._ax = None

    def set_ax(self, cb):
        self._ax = weakref.ref(cb, self._ax_dead)
        self._ax().add_member(self)

    def unset_ax(self):
        if self._ax is not None:
            self._ax().rm_member(self)
            self._ax = None


class YUser(object):
    def __init__(self):
        self._ay = None

    def get_yaxisparam(self):
        if self._ay is None:
            # create colorbar
            figaxes = self.get_figaxes()
            if figaxes is None:
                return None
            if len(figaxes._yaxis) == 0:
                figaxes.add_axis_param(dir='y')
            get_axisparam(figaxes, figaxes._yaxis, self,
                          self.set_ay, self._container_idx)
#           self.set_ay(figaxes._yaxis[0])
        return self._ay()

    def _ay_dead(self, ref):
        self._ay = None

    def set_ay(self, cb):
        self._ay = weakref.ref(cb, self._ay_dead)
        self._ay().add_member(self)

    def unset_ay(self):
        if self._ay is not None:
            self._ay().rm_member(self)
            self._ay = None


class ZUser(object):
    def __init__(self):
        self._az = None

    def get_zaxisparam(self):
        if self._az is None:
            # create colorbar
            figaxes = self.get_figaxes()
            if figaxes is None:
                return None
            if len(figaxes._zaxis) == 0:
                figaxes.add_axis_param(dir='z')
            get_axisparam(figaxes, figaxes._zaxis, self,
                          self.set_az, self._container_idx)
        return self._az()

    def _az_dead(self, ref):
        self._az = None

    def set_az(self, cb):
        self._az = weakref.ref(cb, self._az_dead)
        self._az().add_member(self)

    def unset_az(self):
        if self._az is not None:
            self._az().rm_member(self)
            self._az = None


class CUser(object):
    def __init__(self):
        self._ac = None

    def get_caxisparam(self):
        if self._ac is None:
            # create colorbar
            figaxes = self.get_figaxes()
            if figaxes is None:
                return None
            if len(figaxes._caxis) == 0:
                figaxes.add_axis_param(dir='c')
            get_axisparam(figaxes, figaxes._caxis, self,
                          self.set_ac, self._container_idx)
        return self._ac()

    def get_caxis_choices(self):
        figaxes = self.get_figaxes()
        if figaxes is None:
            return []
        return [c.name for c in figaxes._caxis]

    def set_selected_caxis(self, value, a):
        oc = self._ac()
        figaxes = self.get_figaxes()
        for c in figaxes._caxis:
            if c.name == value:
                break
        self.unset_ac()
        self.set_ac(c)
        if oc is not c:
            c.set_crangeparam_to_artist(a)

    def get_selected_caxis(self, a):
        return self._ac().name

    def _ac_dead(self, ref):
        self._ac = None

    def set_ac(self, cb):
        self._ac = weakref.ref(cb, self._ac_dead)
        self._ac().add_member(self)

    def unset_ac(self):
        if self._ac is not None:
            self._ac().rm_member(self)
            self._ac = None

    def has_cbar(self):
        figaxes = self.get_figaxes()
        if self._ac is None:
            False
        if figaxes is None:
            False
        ac = self._ac()
        return ac.has_cbar()

    def show_cbar(self):
        figaxes = self.get_figaxes()
        if self._ac is None:
            return
        if figaxes is None:
            return
        ac = self._ac()
        ac.show_cbar(figaxes)

    def hide_cbar(self):
        figaxes = self.get_figaxes()
        if self._ac is None:
            return
        if figaxes is None:
            return
        ac = self._ac()
        ac.hide_cbar()

    def onShowCB1(self, event):
        new_cb = False
        if self.has_cbar():
            new_cb = True
        self.show_cbar()
        canvas = event.GetEventObject()
        canvas.draw()
        ifigure.events.SendChangedEvent(self.get_figaxes(),
                                        w=canvas)
        return 1

    def onHideCB1(self, event):
        self.hide_cbar()
        canvas = event.GetEventObject()
        canvas.draw()
        ifigure.events.SendChangedEvent(self.get_figaxes(),
                                        w=canvas)
        return 1



'''
   pid_exits

   based on Q&A in 
   http://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid
 

   however, if access is denied, it returns True. 
'''
import os
import logging


def pid_exists(pid):
    """Check whether pid exists in the current process table."""
    if os.name == 'posix':
        import errno
        if pid < 0:
            return False
        try:
            os.kill(pid, 0)
        except OSError as e:
            return e.errno == errno.EPERM
        else:
            return True
    else:
        import ctypes
        kernel32 = ctypes.windll.kernel32
        HANDLE = ctypes.c_void_p
        DWORD = ctypes.c_ulong
        LPDWORD = ctypes.POINTER(DWORD)

        class ExitCodeProcess(ctypes.Structure):
            _fields_ = [('hProcess', HANDLE),
                        ('lpExitCode', LPDWORD)]
        SYNCHRONIZE = 0x100000
        process = kernel32.OpenProcess(SYNCHRONIZE, 0, pid)

        if not process:
            return False
        ec = ExitCodeProcess()
        out = kernel32.GetExitCodeProcess(process, ctypes.byref(ec))
        if not out:
            err = kernel32.GetLastError()
            if kernel32.GetLastError() == 5:
                # Access is denied.
                logging.warning("Access is denied to get pid info." + str(pid))
            kernel32.CloseHandle(process)
            return True
        elif bool(ec.lpExitCode):
            # print ec.lpExitCode.contents
            # There is an exist code, it quit
            kernel32.CloseHandle(process)
            return False
        # No exit code, it's running.
        kernel32.CloseHandle(process)
        return True



def find_all_keys(d, key):
    return [x for x in list(d) if x.startswith(key+'(')]


def find_idx(keys):
    return [int((x.split('(')[1]).strip()[:-1]) for x in keys]


def key_exists(d, key):
    return key in d or len(find_all_keys(d, key)) > 0


def set_value(d, key, value, size, defv):
    v = [defv]*size
    if len(value) <= len(v):
        v[:len(value)] = value
    else:
        v = value[:len(v)]
    d[key] = v


def clean_key(d, key):
    keys = find_all_keys(d, key)
    for k in keys:
        d.pop(k)
    return d


def read_value(d, key, def_v=None, size=None):
    if key in d:
        return d[key]
    else:
        keys = find_all_keys(d, key)
        idx = find_idx(keys)
        if size is None:
            val = [def_v] * max(idx)
        else:
            val = [def_v] * size

        for i, k in zip(idx, keys):
            val[i-1] = d[k][0]
        return val
